==> syd-3.35.2.tar.gz: syd-v3.35.2/src/kernel/getdents.rs <==
//
// Syd: rock-solid application kernel
// src/kernel/getdents.rs: getdents64(2) handler
//
// Copyright (c) 2023, 2024, 2025 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: GPL-3.0

use std::os::fd::{AsRawFd, RawFd};

use libseccomp::ScmpNotifResp;
use nix::{errno::Errno, NixPath};

use crate::{
    compat::getdents64,
    config::{DIRENT_BUF_SIZE, MMAP_MIN_ADDR, PROC_FILE},
    fs::readlinkat,
    hook::{RemoteProcess, UNotifyEventRequest},
    kernel::sandbox_path,
    path::XPathBuf,
    sandbox::Capability,
};

#[allow(clippy::cognitive_complexity)]
pub(crate) fn sys_getdents64(request: UNotifyEventRequest) -> ScmpNotifResp {
    syscall_handler!(request, |request: UNotifyEventRequest| {
        let req = request.scmpreq;

        if req.data.args[2] == 0 {
            // SAFETY: Result buffer is too small.
            return Err(Errno::EINVAL);
        }

        if req.data.args[1] < *MMAP_MIN_ADDR {
            // SAFETY: If the second argument which must hold a pointer to a
            // linux_dirent structure is not valid, we must return EFAULT
            // without further processing here.
            return Err(Errno::EFAULT);
        }

        // Initial getdents call has Readdir capability,
        // but each file gets listed gets checked for Stat.
        let mut sandbox = request.get_sandbox();
        let hide = sandbox.enabled(Capability::CAP_STAT);
        let safe_name = !sandbox.allow_unsafe_filename();

        // Get remote fd and readlink /proc/self/fd/$fd.
        #[allow(clippy::cast_possible_truncation)]
        let fd = request.get_fd(req.data.args[0] as RawFd)?;
        let mut pfd = XPathBuf::from("self/fd");
        pfd.push_fd(fd.as_raw_fd());

        // Note, the readdir access check here has been
        // moved to the open handler for simplicity and
        // efficiency.
        let mut dir = match readlinkat(PROC_FILE(), &pfd) {
            Ok(dir) if dir.is_relative() => {
                // /proc/1/fd/0 -> pipe:42
                return Err(Errno::ENOTDIR);
            }
            Ok(dir) => dir,
            Err(_) => return Err(Errno::ENOENT),
        };

        // SAFETY: The count argument to the getdents call
        // must not be fully trusted, it can be overly large,
        // and allocating a Vector of that capacity may overflow.
        // This bug was discovered by trinity in this build:
        // https://builds.sr.ht/~alip/job/1077263
        let count = usize::try_from(req.data.args[2])
            .or(Err(Errno::EINVAL))?
            .min(DIRENT_BUF_SIZE);
        let pid = req.pid();
        let len = dir.len();
        let mut dot: u8 = 0;
        let mut ret: u64 = 0;
        while ret == 0 {
            // Release the read-lock before emulation.
            drop(sandbox);

            let mut entries = match getdents64(&fd, count) {
                Ok(entries) => entries,
                Err(Errno::ECANCELED) => break, // EOF or empty directory
                Err(errno) => return Err(errno),
            };

            // Re-acquire the read-lock for access check.
            sandbox = request.get_sandbox();

            #[allow(clippy::arithmetic_side_effects)]
            for entry in &mut entries {
                if dot < 2 && entry.is_dot() {
                    // SAFETY: Allow the special dot entries `.` and `..`.
                    // Note, `..` may point to a denylisted directory,
                    // however at this point there's not much we can do:
                    // even the root directory, ie `/`, has a `..`. In
                    // this exceptional case `..` points to `.`.
                    dot += 1;
                } else {
                    // Append entry name to the directory.
                    dir.push(entry.name_bytes());

                    // SAFETY: Run XPath::check() with file type for global restrictions.
                    if dir
                        .check(
                            pid,
                            Some(&entry.file_type()),
                            Some(entry.as_xpath()),
                            safe_name,
                        )
                        .is_err()
                    {
                        // skip entry.
                        dir.truncate(len);
                        continue;
                    }

                    // Unused when request.is_some()
                    let process = RemoteProcess::new(request.scmpreq.pid());

                    // SAFETY: Run sandbox access check with stat capability.
                    let err = sandbox_path(
                        Some(&request),
                        &sandbox,
                        &process,
                        &dir,
                        Capability::CAP_STAT,
                        hide,
                        "stat",
                    )
                    .is_err();
                    dir.truncate(len);
                    if err {
                        // skip entry.
                        continue;
                    }
                }

                // Access granted, write entry to sandbox process memory.
                match request.write_mem(entry.as_bytes(), req.data.args[1] + ret) {
                    Ok(n) => {
                        ret += n as u64;
                        if n != entry.size() {
                            break;
                        }
                    }
                    Err(_) if ret > 0 => break,
                    Err(errno) => return Err(errno),
                };
            }
        }

        #[allow(clippy::cast_possible_wrap)]
        Ok(request.return_syscall(ret as i64))
    })
}

==> syd-3.36.0.tar.gz: syd-v3.36.0/src/kernel/getdents.rs <==
//
// Syd: rock-solid application kernel
// src/kernel/getdents.rs: getdents64(2) handler
//
// Copyright (c) 2023, 2024, 2025 Ali Polatel <alip@chesswob.org>
//
// SPDX-License-Identifier: GPL-3.0

use std::os::fd::{AsRawFd, RawFd};

use libseccomp::ScmpNotifResp;
use nix::{errno::Errno, NixPath};

use crate::{
    compat::getdents64,
    config::{DIRENT_BUF_SIZE, MMAP_MIN_ADDR, PROC_FILE},
    fs::readlinkat,
    hook::{RemoteProcess, UNotifyEventRequest},
    kernel::sandbox_path,
    path::XPathBuf,
    sandbox::Capability,
};

#[allow(clippy::cognitive_complexity)]
pub(crate) fn sys_getdents64(request: UNotifyEventRequest) -> ScmpNotifResp {
    syscall_handler!(request, |request: UNotifyEventRequest| {
        let req = request.scmpreq;

        // Validate result buffer.
        if req.data.args[2] == 0 {
            return Err(Errno::EINVAL);
        }

        // Validate file descriptor.
        //
        // AT_FDCWD is an invalid file descriptor.
        let fd = RawFd::try_from(req.data.args[0]).or(Err(Errno::EBADF))?;
        if fd < 0 {
            return Err(Errno::EBADF);
        }

        // Validate dirp pointer.
        if req.data.args[1] < *MMAP_MIN_ADDR {
            // SAFETY: If the second argument which must hold a pointer to a
            // linux_dirent structure is not valid, we must return EFAULT
            // without further processing here.
            return Err(Errno::EFAULT);
        }

        // Get remote fd and readlink /proc/self/fd/$fd.
        //
        // Note, the Readdir access check here has been
        // moved to the _open_(2) handler for simplicity and
        // efficiency. The Stat check still takes place.
        let fd = request.get_fd(fd)?;
        let mut pfd = XPathBuf::from("self/fd");
        pfd.push_fd(fd.as_raw_fd());

        let mut dir = match readlinkat(PROC_FILE(), &pfd) {
            Ok(dir) if dir.is_relative() => {
                // /proc/1/fd/0 -> pipe:42
                return Err(Errno::ENOTDIR);
            }
            Ok(dir) => dir,
            Err(_) => return Err(Errno::ENOTDIR),
        };

        // SAFETY: The count argument to the getdents call
        // must not be fully trusted, it can be overly large,
        // and allocating a Vector of that capacity may overflow.
        // This bug was discovered by trinity in this build:
        // https://builds.sr.ht/~alip/job/1077263
        let count = usize::try_from(req.data.args[2])
            .or(Err(Errno::EINVAL))?
            .min(DIRENT_BUF_SIZE);
        let pid = req.pid();
        let len = dir.len();
        let mut dot: u8 = 0;
        let mut ret: u64 = 0;
        while ret == 0 {
            let mut entries = match getdents64(&fd, count) {
                Ok(entries) => entries,
                Err(Errno::ECANCELED) => break, // EOF or empty directory
                Err(errno) => return Err(errno),
            };

            // Lock sandbox for read to perform Stat access check.
            let sandbox = request.get_sandbox();
            let safe_name = !sandbox.allow_unsafe_filename();

            #[allow(clippy::arithmetic_side_effects)]
            for entry in &mut entries {
                if dot < 2 && entry.is_dot() {
                    // SAFETY: Allow the special dot entries `.` and `..`.
                    // Note, `..` may point to a denylisted directory,
                    // however at this point there's not much we can do:
                    // even the root directory, ie `/`, has a `..`. In
                    // this exceptional case `..` points to `.`.
                    dot += 1;
                } else {
                    // Append entry name to the directory.
                    dir.push(entry.name_bytes());

                    // SAFETY: Run XPath::check() with file type for global restrictions.
                    if dir
                        .check(
                            pid,
                            Some(&entry.file_type()),
                            Some(entry.as_xpath()),
                            safe_name,
                        )
                        .is_err()
                    {
                        // skip entry.
                        dir.truncate(len);
                        continue;
                    }

                    // Unused when request.is_some()
                    let process = RemoteProcess::new(request.scmpreq.pid());

                    // SAFETY: Run sandbox access check with stat capability.
                    let err = sandbox_path(
                        Some(&request),
                        &sandbox,
                        &process,
                        &dir,
                        Capability::CAP_STAT,
                        false,
                        "stat",
                    )
                    .is_err();
                    dir.truncate(len);
                    if err {
                        // skip entry.
                        continue;
                    }
                }

                // Access granted, write entry to sandbox process memory.
                match request.write_mem(entry.as_bytes(), req.data.args[1] + ret) {
                    Ok(n) => {
                        ret += n as u64;
                        if n != entry.size() {
                            break;
                        }
                    }
                    Err(_) if ret > 0 => break,
                    Err(errno) => return Err(errno),
                };
            }
        }

        #[allow(clippy::cast_possible_wrap)]
        Ok(request.return_syscall(ret as i64))
    })
}
